# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function

# import os
import sys
import copy
import numpy as np
from numbers import Integral
from collections import OrderedDict
from collections.abc import Sequence

__author__ = "Joerg Neugebauer"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"


class SparseListElement(object):
    """
    Handle single element of a sparse lisr
    Args:
        ind: index
        val: value
    """

    def __init__(self, ind, val):
        self.index = ind
        self.value = val

    def __str__(self):
        return "({}: {})".format(self.index, self.value)


class SparseList(object):
    """
    Object to represent a single sparse list
    Internal representation like a dict
    External representation like a list
    Args:
        sparse_list: dict object with {index: val}
        default: default value for all elements not given by index in sparse_list
        length: length of the list
    """

    def __init__(self, sparse_list, default=None, length=None):
        if isinstance(sparse_list, dict):
            self._dict = sparse_list.copy()
            if "_" in self._dict.keys():
                default = self._dict["_"]
                del self._dict["_"]

            if length is None:
                raise ValueError("Length must be provided in dict input mode")
            self._length = length
        elif isinstance(sparse_list, (list, np.ndarray)):
            # self._dict = {el: [] for el in set(sparse_list)}
            self._dict = {}
            for i, el in enumerate(sparse_list):
                self._dict[i] = el
            self._length = len(sparse_list)
            if length is not None:
                if length != self._length:
                    raise ValueError("Incompatible length of new list")
        self._default = default

    def _val_data_type(self):
        """

        Returns:

        """
        if isinstance(self.values(), dict):
            pass
            print(self.values())
        data_0 = self.values()[0]
        if isinstance(data_0, list):
            if isinstance(data_0[0], bool):
                return "list_bool"
            else:
                raise ValueError(
                    "tags which have as elements lists or tensors are not implemented"
                )
        else:
            return "scalar"

    def to_hdf(self, hdf, key):
        """

        Args:
            hdf:
            key:

        Returns:

        """
        if len(self.list()) > 0:
            # Convert to array and store
            hdf[key] = np.array(self.list())
        elif len(self.values()) > 0:
            print("sparse array: ", key, len(self.values()))
            data_type = self._val_data_type()
            my_dict = OrderedDict()
            my_dict["index"] = self.keys()
            if data_type == "list_bool":
                my_dict["values"] = [
                    sum([2**i * int(v) for i, v in enumerate(val)])
                    for val in self.values()
                ]
            else:
                my_dict["values"] = self.values()
            print("values: ", self.values())
            hdf[key] = my_dict

    def __len__(self):
        return self._length

    def __copy__(self):
        return SparseList(
            sparse_list=self._dict, default=self._default, length=self._length
        )

    def keys(self):
        """

        Returns:
            indices of non-sparse elements
        """
        return self._dict.keys()

    def values(self):
        """

        Returns:
            values of non-sparse elements
        """
        return self._dict.values()

    def items(self):
        """

        Returns:
            index, value pairs of non-sparse elements
        """
        return self._dict.items()

    def list(self):
        """
        convert sparse list into full list
        Returns:
            list representation
        """
        full_list = [self._default for _ in range(self._length)]
        for i, val in self._dict.items():
            full_list[i] = val

        return full_list

    def __iter__(self):
        if self._default is None:
            for i, val in self._dict.items():
                yield SparseListElement(i, val)
        else:
            for i, val in enumerate(self.list()):
                yield val

    def __getitem__(self, item):
        if isinstance(item, Integral):
            if item > len(self):
                raise IndexError
            item = item % len(self)
            if item in self._dict:
                return self._dict[item]
            return self._default

        if isinstance(item, slice):
            ind_list = range(len(self))[item]
        elif isinstance(item, (list, tuple, np.ndarray)):
            if len(item) == 0:
                ind_list = []
            else:
                # for some reason isinstance(True, int) == True, so check for bools first explicitly
                if isinstance(item[0], (bool, np.bool_)):
                    if len(item) != len(self):
                        raise IndexError(
                            "Length of boolean index does not match indexed list!"
                        )
                    ind_list = np.argwhere(item).flatten()
                elif isinstance(item[0], Integral):
                    ind_list = item
        else:
            raise ValueError("Unknown item type: " + str(type(item)))
        sliced_dict = {
            j: self._dict[ind] for j, ind in enumerate(ind_list) if ind in self._dict
        }

        return self.__class__(sliced_dict, default=self._default, length=len(ind_list))

    def __setitem__(self, key, value):
        if isinstance(key, Integral):
            if key > len(self):
                raise IndexError
            key = key % len(self)
            self._dict[key] = value
            return
        elif isinstance(key, slice):
            key = range(len(self))[key]

        if max(key) > self._length:
            raise IndexError

        keys = list(key)
        if isinstance(keys[0], (bool, np.bool_)):
            for i in np.argwhere(keys).flatten():
                self._dict[i] = value
        else:
            for i in key:
                self._dict[i] = value

    def __delitem__(self, key):
        # programmed for simplicity, not for performance
        ind_list = list(range(len(self)))
        if isinstance(key, (list, np.ndarray, tuple)):
            indexes = sorted(list(key), reverse=True)
            for index in indexes:
                del ind_list[index]
        else:
            del ind_list[key]
        new_list = self[ind_list]
        self._dict = new_list._dict
        self._length = new_list._length
        self._default = new_list._default

    def __add__(self, other):
        if not (isinstance(other, SparseList)):
            raise AssertionError()
        if not (self._default == other._default):
            raise AssertionError()
        new_list = self.__copy__()
        shifted_dict = {i + self._length: val for i, val in other._dict.items()}
        new_list._dict.update(shifted_dict)
        new_list._length += len(other)
        return new_list

    def __mul__(self, other):
        if not isinstance(other, Integral):
            raise ValueError("Multiplication defined only for SparseArray*integers")
        overall_list = other * np.arange(len(self)).tolist()
        new_dic = dict()
        for k in self.keys():
            for val in np.argwhere(np.array(overall_list) == k).flatten():
                new_dic[val] = self[k]
        return self.__class__(new_dic, default=self._default, length=other * len(self))

    def __rmul__(self, other):
        if isinstance(other, Integral):
            return self * other

    def __str__(self):
        if self._default is None:
            return "[" + " ".join([str(el) for el in self]) + "]"
        else:
            # return "[" + " ".join([str(el) + os.sep for el in self.list()]) + "]"
            return "[" + " ".join([str(el) for el in self.list()]) + "]"

    def __repr__(self):
        return str(self.list())


def sparse_index(index_list, length, default_val=True):
    """

    Args:
        index_list:
        length:
        default_val:

    Returns:

    """
    new_dict = {i: default_val for i in index_list}
    return SparseList(new_dict, length=length)


class SparseArrayElement(object):
    """
    Single element of a SparseArray
    Args:
        **qwargs:
    """

    def __init__(self, **qwargs):
        self._lists = dict()
        if qwargs:
            self._lists = qwargs

    def __getattr__(self, item):
        if item in self._lists.keys():
            return self._lists[item]
        raise AttributeError(
            "Object has no attribute {} {}".format(self.__class__, item)
        )

    def __str__(self):
        out_str = ""
        for key, val in self._lists.items():
            out_str += "{}: {}".format(key, val)
        return out_str

    def __eq__(self, other):
        if not (isinstance(other, SparseArrayElement)):
            raise AssertionError()
        conditions = []
        for key in self._lists.keys():
            try:
                if isinstance(self._lists[key], np.ndarray):
                    conditions += list(np.equal(self._lists[key], other._lists[key]))
                else:
                    conditions.append(self._lists[key] == other._lists[key])
            except KeyError:
                conditions.append(False)
        return all(conditions)


class SparseArray(object):
    """
    Administrate object that consists of several sparse lists (tags) and full lists that have identical indices and
    length

    Args:
        **qwargs: dictionary containing lists and SparseLists (tags) (must have identical length)
    """

    def __init__(self, length=None, **qwargs):
        self._lists = dict()
        self._length = length
        for key in qwargs:
            value = qwargs[key]
            if self._length is None:
                self._length = len(value)
            else:
                if not len(self) == len(value):
                    raise ValueError(
                        "Inconsistent vector lengths {} {} {}".format(
                            key, len(self), len(value)
                        )
                    )
            self._lists[key] = value

    def __setitem__(self, key, value):
        # exclude hidden variables (starting with _ from being added to _lists
        # if (not hasattr(self, '_lists')) or (key[0] == "_"):
        #     self.__dict__[key] = value
        #     return
        # el

        if isinstance(value, SparseList):
            self._lists[key] = value
            return
        elif isinstance(value, (Sequence, np.ndarray)):
            if len(value) == len(self):
                self._lists[key] = value
                return
            else:
                raise ValueError(
                    "Length of array object and new list are inconsistent: {} {} {}".format(
                        key, len(value), len(self)
                    )
                )
        raise ValueError("Unsupported argument: " + str(type(value)))

    def __getattr__(self, item):
        # if not (item in ["_lists"]):
        #     print "item: ", item, hasattr(self, item)
        if "_lists" in dir(self):  # Python 3
            if item in self._lists.keys():
                return self._lists[item]

        return object.__getattribute__(self, item)
        # raise AttributeError("%r object has no attribute %r" %(self.__class__, item))

    def __delitem__(self, key):
        for k in self.keys():
            if len(self._lists[k]) == 0:
                # ensure ASE compatibility
                print("Empty key in SparseList: ", k, key)
                continue
            # print "del: ", k, key
            if isinstance(self._lists[k], np.ndarray):
                self._lists[k] = np.delete(self._lists[k], key, axis=0)
                self._length = len(self._lists[k])
            elif isinstance(self._lists[k], (list, tuple)):
                if isinstance(key, (list, np.ndarray, tuple)):
                    indexes = sorted(list(key), reverse=True)
                    for index in indexes:
                        del self._lists[k][index]
                else:
                    del self._lists[k][key]
            else:
                del self._lists[k][key]
                # self._length = len(self._lists[k])

    def check_consistency(self):
        """

        Returns:

        """
        for key, val in self._lists.items():
            # for val in self._lists.values():
            #     print ("consistency: ", key, len(val), len(self))
            if not (len(val) == self._length):
                raise AssertionError()

    def __str__(self):
        out_str = "\n"
        for key, val in self._lists.items():
            out_str += key + " := [" + " ".join([str(el) for el in val]) + "] \n"
        return out_str

    def __len__(self):
        if hasattr(self, "_length"):
            return self._length
        else:
            return 0

    def __getitem__(self, item):
        new_dict = {}
        if isinstance(item, Integral):
            for key, value in self._lists.items():
                if value[item] is not None:
                    new_dict[key] = value[item]
            return SparseArrayElement(**new_dict)
        elif isinstance(item, str):
            return self._lists[item]

        elif isinstance(item, (list, np.ndarray)):
            # print("key(__getitem__) len, type, item[0]: ", len(item), type(item), item[0])
            if len(item) == len(self):
                if isinstance(item[0], (np.bool_, bool)):
                    item = np.arange(len(item))[item]
        for key, value in self._lists.items():
            # print ('key: ', key, type(value))
            if isinstance(item, slice):
                new_dict[key] = value[item]
            else:
                if isinstance(value, (list, tuple)):
                    new_dict[key] = [value[i] for i in item]
                else:
                    if len(value) > 0:
                        try:
                            new_dict[key] = value[item]
                        except IndexError:
                            print("Index error:: ", key, item, value)
                    # else:
                    #     new_dict[key] = []
        # print ("new_dict: ", new_dict, self.__class__)
        return self.__class__(**new_dict)

    def keys(self):
        """

        Returns:

        """
        return self._lists.keys()

    def items(self):
        """

        Returns:

        """
        return self._lists.items()

    def __copy__(self):
        """

        Returns:

        """
        cls = self.__class__
        result = cls.__new__(cls)
        result.__init__()
        for k, v in self.__dict__.items():
            if k == "_lists":
                result.__dict__[k] = {}
                for key, val in self._lists.items():
                    if isinstance(val, SparseList):
                        result.__dict__[k][key] = val.__copy__()
                    elif isinstance(val, list):
                        result.__dict__[k][key] = val[:]
                    else:
                        result.__dict__[k][key] = np.copy(val)
            else:
                result.__dict__[k] = v
        return result

    def __add__(self, other):
        # print "__add__.new_elements"
        # assert(isinstance(other, self.__class__))
        new_array = self.__copy__()
        for key, val in other.items():
            if key not in self.keys():
                if isinstance(val, SparseList):
                    new_array._lists[key] = SparseList(
                        {}, default=other._lists[key]._default, length=len(self)
                    )
                else:
                    raise ValueError(
                        "Incompatible lists (for non-sparse lists keys must be identical (1)"
                        + str(key)
                    )

        new_length = len(self) + len(other)
        for key, val in new_array.items():
            # print "key: ", key, val.__class__, isinstance(new_array, SparseList)
            if key in other.keys():
                if isinstance(new_array._lists[key], np.ndarray):
                    new_array._lists[key] = np.append(
                        new_array._lists[key], other._lists[key], axis=0
                    )
                elif isinstance(new_array._lists[key], (list, SparseList)):
                    new_array._lists[key] += other._lists[key]
                else:
                    raise ValueError(
                        "Type not implemented " + str(type(new_array._lists[key]))
                    )
            elif isinstance(val, SparseList):
                new_array._lists[
                    key
                ]._length = new_length  # TODO: default extends to all elements (may be undesired)
            else:
                print("non-matching key: ", key)
                raise ValueError(
                    "Incompatible lists (for non-sparse lists keys must be identical (2)"
                )
        new_array._length += len(other)
        return new_array

    def __mul__(self, other):
        if not isinstance(other, Integral):
            raise ValueError(
                "Multiplication with SparseMatrix only implemented for integers"
            )
        new_array = self.__copy__()
        for key, value in self.items():
            new_array._lists[key] *= other

        new_array._length *= other
        return new_array

    def __rmul__(self, other):
        if isinstance(other, Integral):
            return self * other

    def add_tag(self, *args, **qwargs):
        for key in args:
            self._lists[key] = SparseList({}, length=len(self))

        for key, default in qwargs.items():
            self._lists[key] = SparseList({}, default=default, length=len(self))

    def remove_tag(self, *args, **qwargs):
        """

        Args:
            *args:
            **qwargs:

        Returns:

        """
        for key in args:
            del self._lists[key]
        for key, default in qwargs.items():
            del self._lists[key]



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import random
import warnings
import itertools
from multiprocessing import cpu_count
from pyiron_atomistics.atomistics.job.atomistic import AtomisticGenericJob
from pyiron_base import state, DataContainer, GenericParameters, ImportAlarm
from pyiron_atomistics.atomistics.structure.periodic_table import ChemicalElement
from pyiron_atomistics.atomistics.structure.atoms import (
    Atoms,
    ase_to_pyiron,
    pyiron_to_ase,
)
import numpy as np

try:
    from sqsgenerator.core.sqs import ParallelSqsIterator

    import_alarm = ImportAlarm(
        "SQSJob relies on the [sqsgenerator module](https://github.com/dgehringer/sqsgenerator) module. The package "
        "got a major update (0.0.5 -> 0.1) recently. Please consider to update the package, e. g. with "
        "`conda update -c conda-forge sqsgenerator`"
    )
    DEPRECATED_SQS_API = True
except ImportError:
    DEPRECATED_SQS_API = None

try:
    from sqsgenerator.settings import BadSettings
    from sqsgenerator import sqs_optimize, process_settings, IterationMode

    DEPRECATED_SQS_API = False
    import_alarm = ImportAlarm()
except ImportError:
    import_alarm = ImportAlarm(
        "SQSJob relies on the sqsgenerator module, but this is unavailable. Please ensure your "
        "python environment contains the [sqsgenerator module](https://github.com/dgehringer/sqsgenerator), e.g. with "
        "`conda install -c conda-forge sqsgenerator`."
    )
    DEPRECATED_SQS_API = DEPRECATED_SQS_API or None


from typing import Dict, Optional, Union, Iterable

if DEPRECATED_SQS_API:
    from pymatgen.io.ase import AseAtomsAdaptor

__author__ = "Jan Janssen"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "0.1"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Aug 14, 2020"


def chemical_formula(atoms: Atoms) -> str:
    def group_symbols():
        for species, same in itertools.groupby(atoms.get_chemical_symbols()):
            num_same = len(list(same))
            yield species if num_same == 1 else f"{species}{num_same}"

    return "".join(group_symbols())


def map_dict(f, d: Dict) -> Dict:
    return {k: f(v) for k, v in d.items()}


def mole_fractions_to_composition(
    mole_fractions: Dict[str, float], num_atoms: int
) -> Dict[str, int]:
    # if the sum of x is less than 1 - 1/n then we are missing at least one atoms
    if not (1.0 - 1 / num_atoms) < sum(mole_fractions.values()) < (1.0 + 1 / num_atoms):
        raise ValueError(
            "mole-fractions must sum up to one: {}".format(sum(mole_fractions.values()))
        )

    composition = map_dict(lambda x: x * num_atoms, mole_fractions)
    # check to avoid partial occupation -> x_i * num_atoms is not an integer number
    if any(
        map(
            lambda occupation: not float.is_integer(round(occupation, 1)),
            composition.values(),
        )
    ):
        # at least one of the specified species exhibits fractional occupation, we try to fix it by rounding
        composition_ = map_dict(lambda occ: int(round(occ)), composition)
        warnings.warn(
            "The current mole-fraction specification cannot be applied to {} atoms, "
            "as it would lead to fractional occupation. Hence, I have changed it from "
            '"{}" -> "{}"'.format(
                num_atoms,
                mole_fractions,
                map_dict(lambda n: n / num_atoms, composition_),
            )
        )
        composition = composition_

    # due to rounding errors there might be a difference of one atom
    actual_atoms = sum(composition.values())
    diff = actual_atoms - num_atoms
    if abs(diff) == 1:
        # it is not possible to distribute atoms equally e.g x_a = x_b = x_c = 1/3 on 32 atoms
        # we remove one randomly bet we inform the user
        removed_species = random.choice(tuple(composition))
        composition[removed_species] -= 1
        warnings.warn(
            'It is not possible to distribute the species properly. Therefore one "{}" atom was removed. '
            "This changes the input mole-fraction specification. "
            '"{}" -> "{}"'.format(
                removed_species,
                mole_fractions,
                map_dict(lambda n: n / num_atoms, composition),
            )
        )
    elif abs(diff) > 1:
        # something else is wrong with the mole-fractions input
        raise ValueError(
            "Cannot interpret mole-fraction dict {}".format(mole_fractions)
        )

    return composition


if DEPRECATED_SQS_API:

    def pyiron_to_pymatgen(structure):
        return AseAtomsAdaptor.get_structure(pyiron_to_ase(structure))

    def pymatgen_to_pyiron(structure):
        return ase_to_pyiron(AseAtomsAdaptor.get_atoms(structure))

    def get_sqs_structures(
        structure,
        mole_fractions,
        weights=None,
        objective=0.0,
        iterations=1e6,
        output_structures=10,
        num_threads=None,
    ):
        structure = pyiron_to_pymatgen(structure)
        if not weights:
            weights = {i: 1.0 / i for i in range(1, 7)}
        if not num_threads:
            # Thats default behaviour
            num_threads = cpu_count()

        # preprocess mole-fractions, to print warnings is necessary
        num_atoms = len(structure)
        composition = mole_fractions_to_composition(mole_fractions, num_atoms)
        mole_fractions = map_dict(lambda n: n / num_atoms, composition)

        iterator = ParallelSqsIterator(
            structure, mole_fractions, weights, num_threads=num_threads
        )
        structures, decmp, iter_, cycle_time = iterator.iteration(
            iterations=iterations,
            output_structures=output_structures,
            objective=objective,
        )
        return [pymatgen_to_pyiron(s) for s in structures], decmp, iter_, cycle_time

else:

    def remap_sro(species: Iterable[str], array: np.ndarray):
        # remaps computed short-range order parameters to style of sqsgenerator=v0.0.5
        species = tuple(
            sorted(species, key=lambda abbr: ChemicalElement(abbr).atomic_number)
        )
        return {
            "{}-{}".format(si, sj): array[:, i, j].tolist()
            for (i, si), (j, sj) in itertools.product(
                enumerate(species), enumerate(species)
            )
            if j >= i
        }

    def remap_sqs_results(result):
        # makes new interface compatible with old one
        pyiron_structure = ase_to_pyiron(result["structure"])
        return pyiron_structure, remap_sro(
            set(pyiron_structure.get_chemical_symbols()), result["parameters"]
        )

    def transpose(it):
        return zip(*it)

    def get_sqs_structures(
        structure: Atoms,
        mole_fractions: Dict[str, Union[float, int]],
        weights: Optional[Dict[int, float]] = None,
        objective: Union[float, np.ndarray] = 0.0,
        iterations: Union[float, int] = 1e6,
        output_structures: int = 10,
        mode: str = "random",
        num_threads: Optional[int] = None,
        prefactors: Optional[Union[float, np.ndarray]] = None,
        pair_weights: Optional[np.ndarray] = None,
        rtol: Optional[float] = None,
        atol: Optional[float] = None,
        which: Optional[Iterable[int]] = None,
        shell_distances: Optional[Iterable[int]] = None,
        minimal: Optional[bool] = True,
        similar: Optional[bool] = True,
    ):

        structure = pyiron_to_ase(structure)
        composition = mole_fractions_to_composition(mole_fractions, len(structure))

        settings = dict(
            atol=atol,
            rtol=rtol,
            mode=mode,
            which=which,
            structure=structure,
            prefactors=prefactors,
            shell_weights=weights,
            iterations=iterations,
            composition=composition,
            pair_weights=pair_weights,
            target_objective=objective,
            shell_distances=shell_distances,
            threads_per_rank=num_threads or cpu_count(),
            max_output_configurations=output_structures,
        )
        # not specifying a parameter in settings causes sqsgenerator to choose a "sensible" default,
        # hence we remove all entries with a None value

        results, timings = sqs_optimize(
            {param: value for param, value in settings.items() if value is not None},
            minimal=minimal,
            similar=similar,
            make_structures=True,
            structure_format="ase",
        )

        structures, sro_breakdown = transpose(map(remap_sqs_results, results.values()))
        num_iterations = iterations
        cycle_time = np.average(list(map_dict(np.average, timings).values()))
        return structures, sro_breakdown, num_iterations, cycle_time


class SQSJob(AtomisticGenericJob):
    """
    Produces special quasirandom structures designed to duplicate truly random chemical distributions as well as
    possible while using finite periodic cells.

    A pyiron wrapper for the [SQS code of Dominik Gehringer](https://github.com/dgehringer/sqsgenerator).

    'Structural models used in calculations of properties of substitutionally random $A_{1-x}B_x$ alloys are usually
    constructed by randomly occupying each of the N sites of a periodic cell by A or B. We show that it is possible to
    design ‘‘special quasirandom structures’’ (SQS’s) that mimic for small N (even N=8) the first few, physically most
    relevant radial correlation functions of a perfectly random structure far better than the standard technique does.
    We demonstrate the usefulness of these SQS’s by calculating optical and thermodynamic properties of a number of
    semiconductor alloys in the local-density formalism.'
    From the abstract of Zunger, Wei, Ferreira, and Bernard, Phys. Rev. Lett. 65 (1990) 353,
    DOI:https://doi.org/10.1103/PhysRevLett.65.353

    Input:
        - mole_fractions (dict): Approximate chemical composition for the output structure(s), using chemical symbols
            as the keys and floats as the values. Values should sum to 1, but within reason will be automatically
            adjusted to accommodate the number of atoms in the provided structure (adjustments printed to standard out).
            Vacancies can also be included using the key '0'.
        - weights (list/numpy.ndarray): Specifies the weights of the individual shells. (Default is None, which uses the
            inverse of the shell number for the weight, i.e.
            [1, 0.5, 0.3333333333, 0.25, 0.2, 0.166666667, 0.1428571429].)
        - objective (float): Specifies the value the objective functions. The program tries to reach the specified
            objective function. (Default is 0.)
        - iterations (int): How many iterations to make searching for the most special quasirandom structure using a
            random shuffling procedure. (Default is 1e6)
        - n_output_structures (int): How many different SQS structures to return (in decreasing special
            quasirandomness). (Default is 1.)

    Example:
        Case I: Get SQS for a given mole fraction:

        >>> job = SQSJob(job_name='sqs')
        >>> job.structure = structure
        >>> job.input.mole_fractions = {'Al': 0.8, 'Ni':0.2}
        >>> job.run()

        Case II: Get SQS for a given structure already containing at least 2 elements:

        >>> job = SQSJob(job_name='sqs')
        >>> job.structure = structure
        >>> job.run()

        In Case II, if the mole fractions will be overwritten if you specify the values (like in Case I)
    """

    publication = {
        "sqs": {
            "method": {
                "title": "Special quasirandom structures",
                "author": [
                    "Zunger, A.",
                    "Wei, S.-H.",
                    "Ferreira, L.G.",
                    "Bernard, J.E.",
                ],
                "journal": "Phys. Rev. Lett.",
                "volume": "65",
                "issue": "3",
                "pages": "353",
                "numpages": "0",
                "month": "July",
                "publisher": "American Physical Society",
                "doi": "10.1103/PhysRevLett.65.353",
                "url": "https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.65.353",
            }
        }
    }

    @import_alarm
    def __init__(self, project, job_name):
        super(SQSJob, self).__init__(project, job_name)
        # self.input = InputList(table_name='input')
        self.input = DataContainer(table_name="custom_dict")
        self.input.mole_fractions = dict()
        self.input.weights = None
        self.input.objective = 0.0
        self.input.iterations = 1e6
        self.input.n_output_structures = 1
        self._python_only_job = True
        self._lst_of_struct = []
        self.__hdf_version__ = "0.2.0"

        state.publications.add(self.publication)

    @property
    def list_of_structures(self):
        return self._lst_of_struct

    def validate_ready_to_run(self):
        super(SQSJob, self).validate_ready_to_run()
        if len(self.input.mole_fractions) == 0:
            chem = np.unique(self.structure.get_chemical_symbols(), return_counts=True)
            self.input.mole_fractions = dict(zip(chem[0], chem[1] / np.sum(chem[1])))
        if len(self.input.mole_fractions) == 1:
            raise ValueError("There must be at least two chemical elements")

    def db_entry(self):
        """
        Generate the initial database entry

        Returns:
            (dict): db_dict
        """
        db_dict = super(SQSJob, self).db_entry()
        if self.structure:
            struct_len = len(self.structure)
            mole_fractions = {k: v for k, v in self.input.mole_fractions.items()}
            el_lst = sorted(mole_fractions.keys())
            atom_number_lst = [
                int(np.round(mole_fractions[el] * struct_len)) for el in el_lst
            ]
            db_dict["ChemicalFormula"] = "".join(
                [e + str(n) for e, n in zip(el_lst, atom_number_lst)]
            )
        return db_dict

    def list_structures(self):
        return self._lst_of_struct if self.status.finished else []

    def _get_structure(self, frame=-1, wrap_atoms=True):
        return self.list_structures()[frame]

    def _number_of_structures(self):
        return len(self.list_structures())

    # This function is executed
    def run_static(self):
        self._lst_of_struct, decmp, iterations, cycle_time = get_sqs_structures(
            structure=self.structure,
            mole_fractions={k: v for k, v in self.input.mole_fractions.items()},
            weights=self.input.weights,
            objective=self.input.objective,
            iterations=self.input.iterations,
            output_structures=self.input.n_output_structures,
            num_threads=self.server.cores,
        )
        for i, structure in enumerate(self._lst_of_struct):
            with self.project_hdf5.open("output/structures/structure_" + str(i)) as h5:
                structure.to_hdf(h5)
        with self.project_hdf5.open("output") as h5:
            h5["decmp"] = decmp
            h5["cycle_time"] = cycle_time
            h5["iterations"] = iterations
        self.status.finished = True
        self.project.db.item_update(self._runtime(), self.job_id)

    def to_hdf(self, hdf=None, group_name=None):
        super().to_hdf(hdf=hdf, group_name=group_name)
        self._structure_to_hdf()
        with self.project_hdf5.open("input") as h5in:
            self.input.to_hdf(h5in)

    def from_hdf(self, hdf=None, group_name=None):
        super().from_hdf(hdf=hdf, group_name=group_name)
        self._structure_from_hdf()
        self._backwards_compatible_input_from_hdf()
        with self.project_hdf5.open("output/structures") as hdf5_output:
            structure_names = hdf5_output.list_groups()
        for group in structure_names:
            with self.project_hdf5.open("output/structures/" + group) as hdf5_output:
                self._lst_of_struct.append(Atoms().from_hdf(hdf5_output))

    def _backwards_compatible_input_from_hdf(self):
        if "HDF_VERSION" in self.project_hdf5.list_nodes():
            version = self.project_hdf5["HDF_VERSION"]
        else:
            version = "0.1.0"

        if version == "0.1.0":
            with self.project_hdf5.open("input") as hdf5_input:
                gp = GenericParameters(table_name="custom_dict")
                gp.from_hdf(hdf5_input)
                for k in gp.keys():
                    self.input[k] = gp[k]
        elif version == "0.2.0":
            with self.project_hdf5.open("input") as hdf5_input:
                self.input.from_hdf(hdf5_input)
        else:
            raise ValueError("Cannot handle hdf version: {}".format(version))



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import numpy as np
import scipy.constants as spc
import warnings

__author__ = "Joerg Neugebauer, Sudarsan Surendralal, Jan Janssen"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)


# Conversion factors for transfroming pyiron units to Lammps units (alphabetical)

AMU_TO_G = spc.atomic_mass * spc.kilo
AMU_TO_KG = spc.atomic_mass
ANG_PER_FS_TO_ANG_PER_PS = spc.pico / spc.femto
ANG_PER_FS_TO_BOHR_PER_FS = spc.angstrom / spc.physical_constants["Bohr radius"][0]
ANG_PER_FS_TO_CM_PER_S = (spc.angstrom / spc.femto) / spc.centi
ANG_PER_FS_TO_M_PER_S = spc.angstrom / spc.femto
ANG_TO_BOHR = spc.angstrom / spc.physical_constants["Bohr radius"][0]
ANG_TO_CM = spc.angstrom / spc.centi
ANG_TO_M = spc.angstrom
EL_TO_COUL = spc.elementary_charge
EV_PER_ANG_TO_DYNE = (spc.electron_volt / spc.angstrom) / spc.dyne
EV_PER_ANG_TO_HA_PER_BOHR = (
    spc.physical_constants["electron volt-hartree relationship"][0]
    * spc.physical_constants["Bohr radius"][0]
    / spc.angstrom
)
EV_PER_ANG_TO_KCAL_PER_MOL_ANG = spc.eV / (spc.kilo * spc.calorie / spc.N_A)
EV_PER_ANG_TO_N = spc.electron_volt / spc.angstrom
EV_TO_ERG = spc.electron_volt / spc.erg
EV_TO_HA = spc.physical_constants["electron volt-hartree relationship"][0]
EV_TO_J = spc.electron_volt
EV_TO_KCAL_PER_MOL = spc.eV / (spc.kilo * spc.calorie / spc.N_A)
FS_TO_PS = spc.femto / spc.pico
FS_TO_S = spc.femto
GPA_TO_ATM = spc.giga / spc.atm
GPA_TO_BAR = spc.giga / spc.bar
GPA_TO_BARYE = spc.giga / (spc.micro * spc.bar)  # "barye" = 1e-6 bar
GPA_TO_PA = spc.giga

# Conversions for most of the Lammps units to Pyiron units
# Lammps units source doc: https://lammps.sandia.gov/doc/units.html
# Pyrion units source doc: https://pyiron.readthedocs.io/en/latest/source/faq.html
# At time of writing, not all these conversion factors are used, but may be helpful later.

LAMMPS_UNIT_CONVERSIONS = {
    "metal": {
        "mass": 1.0,
        "distance": 1.0,
        "time": FS_TO_PS,
        "energy": 1.0,
        "velocity": ANG_PER_FS_TO_ANG_PER_PS,
        "force": 1.0,
        "temperature": 1.0,
        "pressure": GPA_TO_BAR,
        "charge": 1.0,
    },
    "si": {
        "mass": AMU_TO_KG,
        "distance": ANG_TO_M,
        "time": FS_TO_S,
        "energy": EV_TO_J,
        "velocity": ANG_PER_FS_TO_M_PER_S,
        "force": EV_PER_ANG_TO_N,
        "temperature": 1.0,
        "pressure": GPA_TO_PA,
        "charge": EL_TO_COUL,
    },
    "cgs": {
        "mass": AMU_TO_G,
        "distance": ANG_TO_CM,
        "time": FS_TO_S,
        "energy": EV_TO_ERG,
        "velocity": ANG_PER_FS_TO_CM_PER_S,
        "force": EV_PER_ANG_TO_DYNE,
        "temperature": 1.0,
        "pressure": GPA_TO_BARYE,
        "charge": 4.8032044e-10,  # In statCoulombs, but these are deprecated and thus not in scipt.constants
    },
    "real": {
        "mass": 1.0,
        "distance": 1.0,
        "time": 1.0,
        "energy": EV_TO_KCAL_PER_MOL,
        "velocity": 1.0,
        "force": EV_PER_ANG_TO_KCAL_PER_MOL_ANG,
        "temperature": 1.0,
        "pressure": GPA_TO_ATM,
        "charge": 1.0,
    },
    "electron": {
        "mass": 1.0,
        "distance": ANG_TO_BOHR,
        "time": 1.0,
        "energy": EV_TO_HA,
        "velocity": ANG_PER_FS_TO_BOHR_PER_FS,
        "force": EV_PER_ANG_TO_HA_PER_BOHR,
        "temperature": 1.0,
        "pressure": GPA_TO_PA,
        "charge": 1.0,
    },
}

# Define conversion for volumes based on distances
for values in LAMMPS_UNIT_CONVERSIONS.values():
    values["volume"] = values["distance"] ** 3
    values["dimensionless_integer_quantity"] = 1


# Hard coded list of all quantities we store in pyiron and the type of quantity it stores (Expand if necessary)
_conversion_dict = dict()
_conversion_dict["distance"] = [
    "positions",
    "cells",
    "unwrapped_positions",
    "mean_unwrapped_positions",
]
_conversion_dict["volume"] = ["volume", "volumes"]
_conversion_dict["pressure"] = ["pressure", "pressures", "mean_pressures"]
_conversion_dict["time"] = ["time"]
_conversion_dict["energy"] = [
    "energy_tot",
    "energy_pot",
    "energy_pot_per_atom",
    "mean_energy_pot",
]
_conversion_dict["temperature"] = ["temperature", "temperatures"]
_conversion_dict["velocity"] = ["velocity", "velocities", "mean_velocities"]
_conversion_dict["mass"] = ["mass"]
_conversion_dict["charge"] = ["charges", "charge"]
_conversion_dict["force"] = ["forces", "force", "mean_forces"]
_conversion_dict["dimensionless_integer_quantity"] = ["steps", "indices"]

# Reverse _conversion_dict
quantity_dict = dict()
for key, val in _conversion_dict.items():
    for v in val:
        quantity_dict[v] = key

# Data type dict to ensure that the units are converted to the proper units
dtype_dict = dict()
for key, val in _conversion_dict.items():
    for v in val:
        # Everything except dimensionless integer quantities are stored as floats
        if key in ["dimensionless_integer_quantity"]:
            dtype_dict[v] = np.int
        else:
            dtype_dict[v] = np.float


class UnitConverter:
    """This is a class to aid conversion of physical quantities between LAMMPS and pyiron units."""

    def __init__(self, units):
        """
        Initialize the class by specifiying the type of lammps units used

        Args:
            units (str): The type of LAMMPS units used (eg. metal, real, cgs, lj, etc.)

        """
        self._units = units
        self._dict = LAMMPS_UNIT_CONVERSIONS[self._units]

    def __getitem__(self, quantity):
        """
        Get quantity from `_dict`

        Args:
            quantity (str): The physical quantity

        Returns:
            float: The conversion factor
        """
        return self._dict[quantity]

    def lammps_to_pyiron(self, quantity):
        """
        Get the conversion factor for a given physical quantity to be converted from LAMMPS to pyiron units

        Args:
            quantity (str): The physical quantity (must be a key in the dictionary `_conversion_dict`)

        Returns:

            float: The conversion factor

        """
        return 1.0 / self[quantity]

    def pyiron_to_lammps(self, quantity):
        """
        Get the conversion factor for a given physical quantity to be converted from pyiron to LAMMPS units

        Args:
            quantity (str): The physical quantity (must be a key in the dictionary `_conversion_dict`)

        Returns:

            float: The conversion factor

        """
        return self[quantity]

    def convert_array_to_pyiron_units(self, array, label):
        """
        Convert a labelled numpy array into pyiron units based on the physical quantity the label corresponds to

        Args:
            array (numpy.ndarray/list): The array to be converted
            label (str): The label of the quantity (must be a key in the dictionary `quantity_dict`)

        Returns:
            ndarray: The array after conversion

        """
        if label in quantity_dict.keys():
            return np.array(
                array * self.lammps_to_pyiron(quantity_dict[label]), dtype_dict[label]
            )
        else:
            warnings.warn(
                message="Warning: Couldn't determine the LAMMPS to pyiron unit conversion type of quantity "
                "{}. Returning un-normalized quantity".format(label)
            )
            return array



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from pyiron_atomistics.atomistics.structure.atoms import Atoms
from pyiron_atomistics.atomistics.master.parallel import ParallelMaster
from pyiron_base import JobGenerator

"""
The StructureListMaster class is a parallel master consisting of a list of structures which are executed in parallel.
"""

__author__ = "Jan Janssen"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2018"


class StructureJobGenerator(JobGenerator):
    """
    JobGenerator for the StructureListMaster - this class implements the functions to generate the parameter list,
    modify the individual jobs according to the parameter list and generate the new job names according to the
    parameter list.
    """

    @property
    def parameter_list(self):
        """

        Returns:
            (list): [[index(int), pyiron_atomistics.atomistics.structure.atoms.Atoms], ...]
        """
        return list(enumerate(self._master.structure_lst))

    @staticmethod
    def job_name(parameter):
        """
        Generate job name for a give set of parameters

        Args:
            parameter: For the StructureListMaster the structures are simply numbered - struct_0, struct_1, ...

        Returns:
            str: job name for the next job
        """
        return "struct_" + str(parameter[0])

    def modify_job(self, job, parameter):
        """
        Modify the next job by setting the structure for the specific parameter

        Args:
            job (GenericJob): next job object to be executed
            parameter: includes the atomistic structure

        Returns:
            GenericJob:
        """
        job.structure = parameter[1]
        return job


class StructureListMaster(ParallelMaster):
    """
    The GenericMaster is the template class for all meta jobs - meaning all jobs which contain multiple other jobs. It
    defines the shared functionality of the different kind of job series.

    Args:
        project (ProjectHDFio): ProjectHDFio instance which points to the HDF5 file the job is stored in
        job_name (str): name of the job, which has to be unique within the project

    Attributes:

        .. attribute:: job_name

            name of the job, which has to be unique within the project

        .. attribute:: status

            execution status of the job, can be one of the following [initialized, appended, created, submitted,
                                                                      running, aborted, collect, suspended, refresh,
                                                                      busy, finished]

        .. attribute:: job_id

            unique id to identify the job in the pyiron database

        .. attribute:: parent_id

            job id of the predecessor job - the job which was executed before the current one in the current job series

        .. attribute:: master_id

            job id of the master job - a meta job which groups a series of jobs, which are executed either in parallel
            or in serial.

        .. attribute:: child_ids

            list of child job ids - only meta jobs have child jobs - jobs which list the meta job as their master

        .. attribute:: project

            Project instance the jobs is located in

        .. attribute:: project_hdf5

            ProjectHDFio instance which points to the HDF5 file the job is stored in

        .. attribute:: job_info_str

            short string to describe the job by it is job_name and job ID - mainly used for logging

        .. attribute:: working_directory

            working directory of the job is executed in - outside the HDF5 file

        .. attribute:: path

            path to the job as a combination of absolute file system path and path within the HDF5 file.

        .. attribute:: version

            Version of the hamiltonian, which is also the version of the executable unless a custom executable is used.

        .. attribute:: executable

            Executable used to run the job - usually the path to an external executable.

        .. attribute:: library_activated

            For job types which offer a Python library pyiron can use the python library instead of an external
            executable.

        .. attribute:: server

            Server object to handle the execution environment for the job.

        .. attribute:: queue_id

            the ID returned from the queuing system - it is most likely not the same as the job ID.

        .. attribute:: logger

            logger object to monitor the external execution and internal pyiron warnings.

        .. attribute:: restart_file_list

            list of files which are used to restart the calculation from these files.

        .. attribute:: job_type

            Job type object with all the available job types: ['ExampleJob', 'SerialMaster', 'ParallelMaster',
                                                               'ScriptJob', 'ListMaster']

        .. attribute:: child_names

            Dictionary matching the child ID to the child job name.
    """

    def __init__(self, project, job_name):
        super(StructureListMaster, self).__init__(project, job_name)
        self.__version__ = "0.0.1"
        self._job_generator = StructureJobGenerator(self)
        self._structure_lst = []

    @property
    def structure_lst(self):
        return self._structure_lst

    @structure_lst.setter
    def structure_lst(self, structure_lst):
        self._structure_lst = structure_lst

    def to_hdf(self, hdf=None, group_name=None):
        """
        Store the StructureListMaster object in the HDF5 File

        Args:
            hdf (ProjectHDFio): HDF5 group object - optional
            group_name (str): HDF5 subgroup name - optional
        """
        super(StructureListMaster, self).to_hdf(hdf=hdf, group_name=group_name)
        with self.project_hdf5.open("input/structures") as hdf5_input:
            for ind, struct in enumerate(self.structure_lst):
                struct.to_hdf(hdf=hdf5_input, group_name="s_" + str(ind))

    def from_hdf(self, hdf=None, group_name=None):
        """
        Restore the StructureListMaster object in the HDF5 File

        Args:
            hdf (ProjectHDFio): HDF5 group object - optional
            group_name (str): HDF5 subgroup name - optional
        """
        super(StructureListMaster, self).from_hdf(hdf=hdf, group_name=group_name)
        with self.project_hdf5.open("input/structures") as hdf5_input:
            self._structure_lst = [
                Atoms().from_hdf(hdf5_input, group_name)
                for group_name in sorted(hdf5_input.list_groups())
            ]

    def collect_output(self):
        """
        Implemented for compatibilty
        """
        pass

    def run_if_interactive_non_modal(self):
        """
        Implemented for compatibilty
        """
        raise TypeError

    def _run_if_busy(self):
        """
        Implemented for compatibilty
        """
        pass



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from pyiron_base import TableJob
from pyiron_atomistics.table.funct import (
    get_incar,
    get_sigma,
    get_total_number_of_atoms,
    get_elements,
    get_convergence_check,
    get_number_of_species,
    get_number_of_ionic_steps,
    get_ismear,
    get_encut,
    get_n_kpts,
    get_n_equ_kpts,
    get_number_of_final_electronic_steps,
    get_majority_species,
    get_job_name,
    get_energy_tot,
    get_energy_pot,
    get_energy_free,
    get_energy_int,
    get_energy_tot_per_atom,
    get_energy_pot_per_atom,
    get_energy_free_per_atom,
    get_energy_int_per_atom,
    get_e_conv_level,
    get_f_states,
    get_e_band,
    get_majority_crystal_structure,
    get_equilibrium_parameters,
    get_structure,
    get_forces,
    get_magnetic_structure,
    get_average_waves,
    get_plane_waves,
    get_ekin_error,
    get_volume,
    get_volume_per_atom,
)


__author__ = "Uday Gajera, Jan Janssen, Joerg Neugebauer"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "0.0.1"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2018"


TableJob._system_function_lst += [
    get_incar,
    get_sigma,
    get_total_number_of_atoms,
    get_elements,
    get_convergence_check,
    get_number_of_species,
    get_number_of_ionic_steps,
    get_ismear,
    get_encut,
    get_n_kpts,
    get_n_equ_kpts,
    get_number_of_final_electronic_steps,
    get_majority_species,
    get_job_name,
    get_energy_tot,
    get_energy_pot,
    get_energy_free,
    get_energy_int,
    get_energy_tot_per_atom,
    get_energy_pot_per_atom,
    get_energy_free_per_atom,
    get_energy_int_per_atom,
    get_e_conv_level,
    get_f_states,
    get_e_band,
    get_majority_crystal_structure,
    get_equilibrium_parameters,
    get_structure,
    get_forces,
    get_magnetic_structure,
    get_average_waves,
    get_plane_waves,
    get_ekin_error,
    get_volume,
    get_volume_per_atom,
]



import os
import numpy as np
from typing import Union, List, Tuple
import pandas as pd

from pyiron_base import DataContainer
from pyiron_atomistics.lammps.potential import LammpsPotential, LammpsPotentialFile
from pyiron_base import GenericJob, ImportAlarm
from pyiron_atomistics.lammps.structure import LammpsStructure, UnfoldingPrism

with ImportAlarm(
    "Calphy functionality requires the `calphy` module (and its dependencies) specified as extra"
    "requirements. Please install it and try again."
) as calphy_alarm:
    from calphy import Calculation, Solid, Liquid, Alchemy
    from calphy.routines import routine_fe, routine_ts, routine_alchemy, routine_pscale

__author__ = "Sarath Menon"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Sarath Menon"
__email__ = "s.menon@mpie.de"
__status__ = "development"
__date__ = "April 1, 2022"

inputdict = {
    "mode": None,
    "pressure": None,
    "temperature": None,
    "reference_phase": None,
    "npt": None,
    "n_equilibration_steps": 15000,
    "n_switching_steps": 25000,
    "n_print_steps": 0,
    "n_iterations": 1,
    "md": {
        "timestep": 0.001,
        "n_small_steps": 10000,
        "n_every_steps": 10,
        "n_repeat_steps": 10,
        "n_cycles": 100,
        "thermostat_damping": 0.5,
        "barostat_damping": 0.1,
    },
    "tolerance": {
        "lattice_constant": 0.0002,
        "spring_constant": 0.01,
        "solid_fraction": 0.7,
        "liquid_fraction": 0.05,
        "pressure": 0.5,
    },
}


class Calphy(GenericJob):
    """
    Class to set up and run calphy jobs for calculation of free energies using LAMMPS.

    An input structure (:attr:`structure`) and interatomic potential (:attr:`potential`) are necessary input options. The additional input options such as the temperature and pressure are specified in the :meth:`.calc_free_energy` method. Depending on the input parameters, a corresponding calculation mode is selected. Further input options can be accessed through :attr:`input.md` and :attr:`input.tolerance`.

    An example which calculates the free energy of Cu using an interatomic potential:

    ```python
    job.structure = pr.create.structure.ase.bulk('Cu', cubic=True).repeat(5)
    job.potential = "2001--Mishin-Y--Cu-1--LAMMPS--ipr1"
    job.calc_free_energy(temperature=1100, pressure=0, reference_phase="solid")
    job.run()
    ```

    In order to calculate the free energy of the liquid phase, the `reference_phase` should be set to `liquid`.

    The different modes can be selected as follows:

    For free energy at a given temperature and pressure:

    ```python
    job.calc_free_energy(temperature=1100, pressure=0, reference_phase="solid")
    ```

    Alternatively, :func:`calc_mode_fe` can be used.

    To obtain the free energy between a given temperature range (temperature scaling):

    ```python
    job.calc_free_energy(temperature=[1100, 1400], pressure=0, reference_phase="solid")
    ```

    Alternatively, :func:`calc_mode_ts` can be used.

    For free energy between a given pressure range (pressure scaling)

    ```python
    job.calc_free_energy(temperature=1000, pressure=[0, 100000], reference_phase="solid")
    ```

    Alternatively, :func:`calc_mode_pscale` can be used.

    To obtain the free energy difference between two interatomic potentials (alchemy/upsampling)

    ```python
    job.potential = ["2001--Mishin-Y--Cu-1--LAMMPS--ipr1", "1986--Foiles-S-M--Cu--LAMMPS--ipr1"]
    job.calc_free_energy(temperature=1100, pressure=0, reference_phase="solid")
    job.run()
    ```

    Alternatively, :func:`calc_mode_alchemy` can be used.

    The way `pressure` is specified determines how the barostat affects the system. For isotropic pressure control:

    ```python
    job.calc_free_energy(temperature=[1100, 1400], pressure=0, reference_phase="solid")
    ```

    For anisotropic pressure control:

    ```python
    job.calc_free_energy(temperature=[1100, 1400], pressure=[0, 0, 0], reference_phase="solid")
    ```

    To constrain the lattice:

    ```python
    job.calc_free_energy(temperature=[1100, 1400], pressure=None, reference_phase="solid")
    ```

    In addition the boolean option :attr:`input.npt` can be used to determine the MD ensemble. If True, temperature integration and alchemy/upsampling are carried out in the NPT ensemble. If False, the NVT ensemble is employed.

    After the calculation is over, the various output options can be accessed through `job.output`.
    """

    def __init__(self, project, job_name):
        super().__init__(project, job_name)
        self.input = DataContainer(inputdict, table_name="inputdata")
        self._potential_initial = None
        self._potential_final = None
        self.input.potential_initial_name = None
        self.input.potential_final_name = None
        self.input.structure = None
        self.output = DataContainer(table_name="output")
        self._data = None
        self.input._pot_dict_initial = None
        self.input._pot_dict_final = None

    def set_potentials(self, potential_filenames: Union[list, str]):
        """
        Set the interatomic potential from a given name

        Args:
            potential_filenames (list, str): list of filenames

        Returns:
            None
        """
        if not isinstance(potential_filenames, list):
            potential_filenames = [potential_filenames]
        if len(potential_filenames) > 0:
            if isinstance(potential_filenames[0], pd.DataFrame):
                potential = potential_filenames[0]
                self.input._pot_dict_initial = potential  # .to_dict()
            else:
                potential = LammpsPotentialFile().find_by_name(potential_filenames[0])
                self.input.potential_initial_name = potential_filenames[0]
            self._potential_initial = LammpsPotential()
            self._potential_initial.df = potential

        if len(potential_filenames) > 1:
            if isinstance(potential_filenames[1], pd.DataFrame):
                potential = potential_filenames[1]
                self.input._pot_dict_final = potential  # .to_dict()
            else:
                potential = LammpsPotentialFile().find_by_name(potential_filenames[1])
                self.input.potential_final_name = potential_filenames[1]
            self._potential_final = LammpsPotential()
            self._potential_final.df = potential
        if len(potential_filenames) > 2:
            raise ValueError("Maximum two potentials can be provided")

    def get_potentials(self) -> List[str]:
        """
        Return the interatomic potentials

        Args:
            None

        Returns:
            list of str: list of interatomic potentials
        """
        if self._potential_final is None:
            return [self._potential_initial.df]
        else:
            return [self._potential_initial.df, self._potential_final.df]

    def copy_pot_files(self):
        """
        Copy potential files to the working directory

        Args:
            None

        Returns:
            None
        """
        if self._potential_initial is not None:
            self._potential_initial.copy_pot_files(self.working_directory)
        if self._potential_final is not None:
            self._potential_final.copy_pot_files(self.working_directory)

    def _prepare_pair_styles(self) -> Tuple[List, List]:
        """
        Prepare pair style and pair coeff

        Args:
            None

        Returns:
            list: pair style and pair coeff
        """
        pair_style = []
        pair_coeff = []
        if self._potential_initial is not None:
            pair_style.append(
                self._potential_initial.df["Config"]
                .to_list()[0][0]
                .strip()
                .split()[1:][0]
            )
            pair_coeff.append(
                " ".join(
                    self._potential_initial.df["Config"]
                    .to_list()[0][1]
                    .strip()
                    .split()[1:]
                )
            )
        if self._potential_final is not None:
            pair_style.append(
                self._potential_final.df["Config"]
                .to_list()[0][0]
                .strip()
                .split()[1:][0]
            )
            pair_coeff.append(
                " ".join(
                    self._potential_final.df["Config"]
                    .to_list()[0][1]
                    .strip()
                    .split()[1:]
                )
            )
        return pair_style, pair_coeff

    def _potential_from_hdf(self):
        """
        Recreate the potential from filename stored in hdf5

        Args:
            None

        Returns:
            None
        """
        filenames = []
        if self.input.potential_initial_name is not None:
            filenames.append(self.input.potential_initial_name)
        elif self.input._pot_dict_initial is not None:
            filenames.append(pd.DataFrame(data=self.input._pot_dict_initial))
        if self.input.potential_final_name is not None:
            filenames.append(self.input.potential_final_name)
        elif self.input._pot_dict_final is not None:
            filenames.append(pd.DataFrame(data=self.input._pot_dict_final))

        self.set_potentials(filenames)

    @property
    def potential(self):
        potentials = self.get_potentials()
        if len(potentials) == 1:
            return potentials[0]
        return potentials

    @potential.setter
    def potential(self, potential_filenames):
        self.set_potentials(potential_filenames)

    @property
    def structure(self):
        return self.input.structure

    @structure.setter
    def structure(self, val):
        self.input.structure = val

    def view_potentials(self) -> List:
        """
        View a list of available interatomic potentials

        Args:
            None

        Returns:
            list: list of available potentials
        """
        if not self.structure:
            raise ValueError("please assign a structure first")
        else:
            list_of_elements = set(self.structure.get_chemical_symbols())
        list_of_potentials = LammpsPotentialFile().find(list_of_elements)
        if list_of_potentials is not None:
            return list_of_potentials
        else:
            raise TypeError(
                "No potentials found for this kind of structure: ",
                str(list_of_elements),
            )

    def list_potentials(self):
        """
        List of interatomic potentials suitable for the current atomic structure.

        use self.potentials_view() to get more details.

        Args:
            None

        Returns:
            list: potential names
        """
        return list(self.view_potentials()["Name"].values)

    def structure_to_lammps(self):
        """
        Convert structure to LAMMPS structure

        Args:
            None

        Returns:
            list: pair style and pair coeff
        """
        prism = UnfoldingPrism(self.input.structure.cell)
        lammps_structure = self.input.structure.copy()
        lammps_structure.set_cell(prism.A)
        lammps_structure.positions = np.matmul(self.input.structure.positions, prism.R)
        return lammps_structure

    def write_structure(self, structure, file_name: str, working_directory: str):
        """
        Write structure to file

        Args:
            structure: input structure
            file_name (str): output file name
            working_directory (str): output working directory

        Returns:
            None
        """
        lmp_structure = LammpsStructure()
        lmp_structure.potential = self._potential_initial
        lmp_structure.el_eam_lst = set(structure.get_chemical_symbols())
        lmp_structure.structure = self.structure_to_lammps()
        lmp_structure.write_file(file_name=file_name, cwd=working_directory)

    def determine_mode(self):
        """
        Determine the calculation mode

        Args:
            None

        Returns:
            None
        """
        if len(self.get_potentials()) == 2:
            self.input.mode = "alchemy"
            self.input.reference_phase = "alchemy"
        elif isinstance(self.input.pressure, list):
            if len(self.input.pressure) == 2:
                self.input.mode = "pscale"
        elif isinstance(self.input.temperature, list):
            if len(self.input.temperature) == 2:
                self.input.mode = "ts"
        else:
            self.input.mode = "fe"
        # if mode was not set, raise Error
        if self.input.mode is None:
            raise RuntimeError("Could not determine the mode")

    def write_input(self):
        """
        Write input for calphy calculation

        Args:
            None

        Returns:
            None
        """
        # now prepare the calculation
        calc = Calculation()
        for key in inputdict.keys():
            if key not in ["md", "tolerance"]:
                setattr(calc, key, self.input[key])
        for key in inputdict["md"].keys():
            setattr(calc.md, key, self.input["md"][key])
        for key in inputdict["tolerance"].keys():
            setattr(calc.tolerance, key, self.input["tolerance"][key])

        file_name = "conf.data"
        self.write_structure(self.structure, file_name, self.working_directory)
        calc.lattice = os.path.join(self.working_directory, "conf.data")

        self.copy_pot_files()
        pair_style, pair_coeff = self._prepare_pair_styles()
        calc._fix_potential_path = False
        calc.pair_style = pair_style
        calc.pair_coeff = pair_coeff

        calc.element = list(np.unique(self.structure.get_chemical_symbols()))
        calc.mass = list(np.unique(self.structure.get_masses()))

        calc.queue.cores = self.server.cores
        self.calc = calc

    def calc_mode_fe(
        self,
        temperature: float = None,
        pressure: Union[list, float, None] = None,
        reference_phase: str = None,
        n_equilibration_steps: int = 15000,
        n_switching_steps: int = 25000,
        n_print_steps: int = 0,
        n_iterations: int = 1,
    ):
        """
        Calculate free energy at given conditions

        Args:
            None

        Returns:
            None
        """
        if temperature is None:
            raise ValueError("provide a temperature")
        if reference_phase is None:
            raise ValueError("provide a reference_phase")

        self.input.temperature = temperature
        self.input.pressure = pressure
        self.input.reference_phase = reference_phase
        self.input.n_equilibration_steps = n_equilibration_steps
        self.input.n_switching_steps = n_switching_steps
        self.input.n_print_steps = n_print_steps
        self.input.n_iterations = n_iterations
        self.input.mode = "fe"

    def calc_mode_ts(
        self,
        temperature: float = None,
        pressure: Union[list, float, None] = None,
        reference_phase: str = None,
        n_equilibration_steps: int = 15000,
        n_switching_steps: int = 25000,
        n_print_steps: int = 0,
        n_iterations: int = 1,
    ):
        """
        Calculate free energy between given temperatures

        Args:
            None

        Returns:
            None
        """
        if temperature is None:
            raise ValueError("provide a temperature")
        if reference_phase is None:
            raise ValueError("provide a reference_phase")

        self.input.temperature = temperature
        self.input.pressure = pressure
        self.input.reference_phase = reference_phase
        self.input.n_equilibration_steps = n_equilibration_steps
        self.input.n_switching_steps = n_switching_steps
        self.input.n_print_steps = n_print_steps
        self.input.n_iterations = n_iterations
        self.input.mode = "ts"

    def calc_mode_alchemy(
        self,
        temperature: float = None,
        pressure: Union[list, float, None] = None,
        reference_phase: str = None,
        n_equilibration_steps: int = 15000,
        n_switching_steps: int = 25000,
        n_print_steps: int = 0,
        n_iterations: int = 1,
    ):
        """
        Perform upsampling/alchemy between two interatomic potentials

        Args:
            None

        Returns:
            None
        """
        if temperature is None:
            raise ValueError("provide a temperature")
        self.input.temperature = temperature
        self.input.pressure = pressure
        self.input.reference_phase = reference_phase
        self.input.n_equilibration_steps = n_equilibration_steps
        self.input.n_switching_steps = n_switching_steps
        self.input.n_print_steps = n_print_steps
        self.input.n_iterations = n_iterations
        self.input.mode = "alchemy"

    def calc_mode_pscale(
        self,
        temperature: float = None,
        pressure: Union[list, float, None] = None,
        reference_phase: str = None,
        n_equilibration_steps: int = 15000,
        n_switching_steps: int = 25000,
        n_print_steps: int = 0,
        n_iterations: int = 1,
    ):
        """
        Calculate free energy between two given pressures

        Args:
            None

        Returns:
            None
        """
        if temperature is None:
            raise ValueError("provide a temperature")
        if reference_phase is None:
            raise ValueError("provide a reference_phase")

        self.input.temperature = temperature
        self.input.pressure = pressure
        self.input.reference_phase = reference_phase
        self.input.n_equilibration_steps = n_equilibration_steps
        self.input.n_switching_steps = n_switching_steps
        self.input.n_print_steps = n_print_steps
        self.input.n_iterations = n_iterations
        self.input.mode = "pscale"

    def calc_free_energy(
        self,
        temperature: float = None,
        pressure: Union[list, float, None] = None,
        reference_phase: str = None,
        n_equilibration_steps: int = 15000,
        n_switching_steps: int = 25000,
        n_print_steps: int = 0,
        n_iterations: int = 1,
    ):
        """
        Calculate free energy at given conditions

        Args:
            None

        Returns:
            None
        """
        if temperature is None:
            raise ValueError("provide a temperature")
        self.input.temperature = temperature
        self.input.pressure = pressure
        self.input.reference_phase = reference_phase
        self.input.n_equilibration_steps = n_equilibration_steps
        self.input.n_switching_steps = n_switching_steps
        self.input.n_print_steps = n_print_steps
        self.input.n_iterations = n_iterations
        self.determine_mode()
        if self.input.mode != "alchemy":
            if reference_phase is None:
                raise ValueError("provide a reference_phase")

    def run_static(self):
        self.status.running = True
        if self.input.reference_phase == "alchemy":
            job = Alchemy(calculation=self.calc, simfolder=self.working_directory)
        elif self.input.reference_phase == "solid":
            job = Solid(calculation=self.calc, simfolder=self.working_directory)
        elif self.input.reference_phase == "liquid":
            job = Liquid(calculation=self.calc, simfolder=self.working_directory)
        else:
            raise ValueError("Unknown reference state")

        if self.input.mode == "alchemy":
            routine_alchemy(job)
        elif self.input.mode == "fe":
            routine_fe(job)
        elif self.input.mode == "ts":
            routine_ts(job)
        elif self.input.mode == "pscale":
            routine_pscale(job)
        else:
            raise ValueError("Unknown mode")
        self._data = job.report
        # save conc for later use
        self.input.concentration = job.concentration
        del self._data["input"]
        self.status.collect = True
        self.run()

    def collect_general_output(self):
        """
        Collect the output from calphy

        Args:
            None

        Returns:
            None
        """
        if self._data is not None:
            if "spring_constant" in self._data["average"].keys():
                self.output.spring_constant = self._data["average"]["spring_constant"]
            self.output.energy_free = self._data["results"]["free_energy"]
            self.output.energy_free_error = self._data["results"]["error"]
            self.output.energy_free_reference = self._data["results"][
                "reference_system"
            ]
            self.output.energy_work = self._data["results"]["work"]
            self.output.temperature = self.input.temperature
            f_ediff, b_ediff, flambda, blambda = self.collect_ediff()
            self.output.forward_energy_diff = list(f_ediff)
            self.output.backward_energy_diff = list(b_ediff)
            self.output.forward_lambda = list(flambda)
            self.output.backward_lambda = list(blambda)
            if self.input.mode == "ts":
                datfile = os.path.join(self.working_directory, "temperature_sweep.dat")
                t, fe, ferr = np.loadtxt(datfile, unpack=True, usecols=(0, 1, 2))
                self.output.energy_free = np.array(fe)
                self.output.energy_free_error = np.array(ferr)
                self.output.temperature = np.array(t)

            elif self.input.mode == "pscale":
                datfile = os.path.join(self.working_directory, "pressure_sweep.dat")
                p, fe, ferr = np.loadtxt(datfile, unpack=True, usecols=(0, 1, 2))
                self.output.energy_free = np.array(fe)
                self.output.energy_free_error = np.array(ferr)
                self.output.pressure = np.array(p)

    def collect_ediff(self):
        """
        Calculate the energy difference between reference system and system of interest

        Args:
            None

        Returns:
            None
        """
        f_ediff = []
        b_ediff = []

        for i in range(1, self.input.n_iterations + 1):
            fwdfilename = os.path.join(self.working_directory, "forward_%d.dat" % i)
            bkdfilename = os.path.join(self.working_directory, "backward_%d.dat" % i)
            nelements = self.calc.n_elements

            if self.input.reference_phase == "solid":
                fdui = np.loadtxt(fwdfilename, unpack=True, comments="#", usecols=(0,))
                bdui = np.loadtxt(bkdfilename, unpack=True, comments="#", usecols=(0,))

                fdur = np.zeros(len(fdui))
                bdur = np.zeros(len(bdui))

                for i in range(nelements):
                    fdur += self.input.concentration[i] * np.loadtxt(
                        fwdfilename, unpack=True, comments="#", usecols=(i + 1,)
                    )
                    bdur += self.input.concentration[i] * np.loadtxt(
                        bkdfilename, unpack=True, comments="#", usecols=(i + 1,)
                    )

                flambda = np.loadtxt(
                    fwdfilename, unpack=True, comments="#", usecols=(nelements + 1,)
                )
                blambda = np.loadtxt(
                    bkdfilename, unpack=True, comments="#", usecols=(nelements + 1,)
                )
            else:
                fdui, fdur, flambda = np.loadtxt(
                    fwdfilename, unpack=True, comments="#", usecols=(0, 1, 2)
                )
                bdui, bdur, blambda = np.loadtxt(
                    bkdfilename, unpack=True, comments="#", usecols=(0, 1, 2)
                )

            f_ediff.append(fdui - fdur)
            b_ediff.append(bdui - bdur)

        return f_ediff, b_ediff, flambda, blambda

    def collect_output(self):
        self.collect_general_output()
        self.to_hdf()

    def to_hdf(self, hdf=None, group_name=None):
        super().to_hdf(hdf=hdf, group_name=group_name)
        self.input.to_hdf(self.project_hdf5)
        self.output.to_hdf(self.project_hdf5)

    def from_hdf(self, hdf=None, group_name=None):
        super().from_hdf(hdf=hdf, group_name=group_name)
        self.input.from_hdf(self.project_hdf5)
        self.output.from_hdf(self.project_hdf5)
        self._potential_from_hdf()

    @property
    def publication(self):
        return {
            "calphy": {
                "calphy": {
                    "title": "Automated free-energy calculation from atomistic simulations",
                    "journal": "Physical Review Materials",
                    "volume": "5",
                    "number": "10",
                    "pages": "103801",
                    "year": "2021",
                    "doi": "10.1103/PhysRevMaterials.5.103801",
                    "url": "https://journals.aps.org/prmaterials/abstract/10.1103/PhysRevMaterials.5.103801",
                    "author": [
                        "Menon, Sarath and Lysogorskiy, Yury and Rogal, Jutta and Drautz, Ralf"
                    ],
                }
            }
        }



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import numpy as np
from pyiron_atomistics.atomistics.job.interactivewrapper import InteractiveWrapper
from pyiron_base import DataContainer
from pyiron_atomistics.atomistics.job.interactive import GenericInteractiveOutput
from scipy.optimize import minimize
import scipy
import warnings

__author__ = "Osamu Waseda"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Osamu Waseda"
__email__ = "waseda@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2018"

GPa_to_eV_by_A3 = (
    1e21 / scipy.constants.physical_constants["joule-electron volt relationship"][0]
)


class ScipyMinimizer(InteractiveWrapper):
    """
    Structure optimization class based on Scipy minimizers.

    Example I:

    # Position optimization
    >>> pr = Project('position')
    >>> job = pr.create_job('SomeAtomisticJob', 'atomistic')
    >>> job.structure = pr.create_structure('Al', 'fcc', 4.)
    >>> # it works also in the static mode, but interactive is recommended
    >>> job.server.run_mode.interactive = True
    >>> minim = pr.create_job('ScipyMinimizer', 'scipy')
    >>> minim.ref_job = job
    >>> minim.run()

    Example II:

    # Volume optimization:
    >>> pr = Project('volume')
    >>> job = pr.create_job('SomeAtomisticJob', 'atomistic')
    >>> job.structure = pr.create_structure('Al', 'fcc', 4.)
    >>> # it works also in the static mode, but interactive is recommended
    >>> job.server.run_mode.interactive = True
    >>> minim = pr.create_job('ScipyMinimizer', 'scipy')
    >>> minim.ref_job = job
    >>> minim.calc_minimize(pressure=0, volume_only=True)
    >>> minim.run()

    By setting `volume_only`, positions are not updated, so that only the
    pressures are minimized.

    It is possible to optimize both the volume and the positions, but since
    changing the cell also changes the definition of coordinates, it is a
    mathematically ill-defined problem and therefore it might end up in a
    physically undefined state. For this reason, it is strongly recommended to
    launch several jobs using the Murnaghan class (cf. `Murnaghan`).

    In order to perform volume optimization, the child job must have
    3x3-pressure output.
    """

    def __init__(self, project, job_name):
        super(ScipyMinimizer, self).__init__(project, job_name)

        self._ref_job = None
        self.input = Input()
        self.output = ScipyMinimizerOutput(job=self)
        self.interactive_cache = {}
        self._delete_existing_job = True

    def set_input_to_read_only(self):
        """
        This function enforces read-only mode for the input classes, but it
        has to be implement in the individual classes.
        """
        super(ScipyMinimizer, self).set_input_to_read_only()
        self.input.read_only = True

    def write_input(self):
        pass

    def _initialize_structure(self):
        self._original_cell = self.ref_job.structure.cell.copy()
        self._current_strain = np.zeros(6)

    def run_static(self):
        self.ref_job_initialize()
        self._logger.debug("cg status: " + str(self.status))
        self._initialize_structure()
        if self.ref_job.server.run_mode.interactive:
            self._delete_existing_job = False
        self.ref_job.run(delete_existing_job=self._delete_existing_job)
        self.status.running = True
        if self.input.pressure is not None:
            x0 = np.zeros(sum(self.input.pressure != None))
            if not self.input.volume_only:
                x0 = np.append(
                    x0, self.ref_job.structure.get_scaled_positions().flatten()
                )
        else:
            x0 = self.ref_job.structure.positions.flatten()
        self.output._result = minimize(
            method=self.input.minimizer,
            fun=self._get_value,
            x0=x0,
            jac=self._get_gradient,
            tol=1.0e-20,
            options={"maxiter": self.input.ionic_steps, "return_all": True},
        )
        self.status.collect = True
        self.collect_output()
        if self.ref_job.server.run_mode.interactive:
            self.ref_job.interactive_close()
        if self["output/convergence"] > 0:
            self.status.finished = True
        else:
            self.status.not_converged = True

    @staticmethod
    def _tensor_to_voigt(s, strain=False):
        ss = 0.5 * (s + s.T)
        ss = ss.flatten()[[0, 4, 8, 5, 2, 1]]
        if strain:
            ss[3:] *= 2
        return ss

    @staticmethod
    def _voigt_to_tensor(s, strain=False):
        ss = np.array(s).copy()
        if not strain:
            ss[:3] /= 2
        ss = np.array([[ss[0], ss[5], ss[4]], [0, ss[1], ss[3]], [0, 0, ss[2]]])
        ss += ss.T
        return ss

    def _update(self, x):
        rerun = False
        if self.input.pressure is not None:
            if not np.allclose(x[: len(self.input.pressure)], self._current_strain):
                if len(self.input.pressure) == 1:
                    self._current_strain[:3] = x[0]
                else:
                    self._current_strain[self.input.pressure != None] = x[
                        : len(self.input.pressure)
                    ]
                cell = np.matmul(
                    self._voigt_to_tensor(self._current_strain, strain=True)
                    + np.eye(3),
                    self._original_cell,
                )
                self.ref_job.structure.set_cell(cell, scale_atoms=True)
                rerun = True
            if not self.input.volume_only and not np.allclose(
                x[len(self.input.pressure) :],
                self.ref_job.structure.get_scaled_positions().flatten(),
            ):
                self.ref_job.structure.set_scaled_positions(
                    x[len(self.input.pressure) :].reshape(-1, 3)
                )
                rerun = True
        elif not np.allclose(x, self.ref_job.structure.positions.flatten()):
            self.ref_job.structure.positions = x.reshape(-1, 3)
            rerun = True
        if rerun:
            self.ref_job.run(delete_existing_job=self._delete_existing_job)

    def check_convergence(self):
        if self.input.ionic_energy_tolerance > 0:
            if len(self.ref_job.output.energy_pot) < 2:
                return False
            elif (
                np.absolute(np.diff(self.ref_job.output.energy_pot)[-1])
                > self.input.ionic_energy_tolerance
            ):
                return False
            if self.input.ionic_force_tolerance == 0:
                return True
        max_force = np.linalg.norm(self.ref_job.output.forces[-1], axis=-1).max()
        if self.input.pressure is None:
            if max_force > self.input.ionic_force_tolerance:
                return False
        elif self.input.volume_only:
            if (
                np.absolute(self._get_pressure() - self.input.pressure).max()
                > self.input.pressure_tolerance
            ):
                return False
        else:
            if max_force > self.input.ionic_force_tolerance:
                return False
            if (
                np.absolute(self._get_pressure() - self.input.pressure).max()
                > self.input.pressure_tolerance
            ):
                return False
        return True

    def _get_pressure(self):
        if len(self.input.pressure) == 1:
            return [np.mean(np.diagonal(self.ref_job.output.pressures[-1]))]
        else:
            return self._tensor_to_voigt(self.ref_job.output.pressures[-1])[
                self.input.pressure != None
            ]

    def _get_gradient(self, x):
        self._update(x)
        prefactor = 1.0e-1
        if self.check_convergence():
            prefactor = 0
        if self.input.pressure is not None:
            pressure = -(self._get_pressure() - self.input.pressure)
            if self.input.volume_only:
                return pressure * prefactor
            else:
                return (
                    np.append(
                        pressure,
                        -np.einsum(
                            "ij,ni->nj",
                            np.linalg.inv(self.ref_job.structure.cell),
                            self.ref_job.output.forces[-1],
                        ).flatten(),
                    ).flatten()
                    * prefactor
                )
        else:
            return -self.ref_job.output.forces[-1].flatten() * prefactor

    def _get_value(self, x):
        self._update(x)
        return self.ref_job.output.energy_pot[-1]

    def collect_output(self):
        self.output.to_hdf(self._hdf5)

    def to_hdf(self, hdf=None, group_name=None):
        super(ScipyMinimizer, self).to_hdf(hdf=hdf, group_name=group_name)
        self.output.to_hdf(self._hdf5)

    def calc_minimize(
        self,
        max_iter=100,
        pressure=None,
        algorithm="CG",
        ionic_energy_tolerance=0,
        ionic_force_tolerance=1.0e-2,
        pressure_tolerance=1.0e-3,
        volume_only=False,
    ):
        """
        Args:
            algorithm (str): scipy algorithm (currently only 'CG' and 'BFGS' run reliably)
            pressure (float/list/numpy.ndarray): target pressures
            max_iter (int): maximum number of iterations
            ionic_energy_tolerance (float): convergence goal in terms of
                                  energy (optional)
            ionic_force_tolerance (float): convergence goal in terms of
                                  forces (optional)
            volume_only (bool): Only pressure minimization
        """
        if pressure is None and volume_only:
            raise ValueError("pressure must be specified if volume_only")
        if pressure is not None and not volume_only:
            warnings.warn(
                "Simultaneous optimization of pressures and positions is a"
                + " mathematically ill posed problem - there is no guarantee"
                + " that it converges to the desired structure"
            )
        if pressure is not None:
            pressure = np.array([pressure]).flatten()
            if len(pressure) == 9:
                pressure = self._tensor_to_voigt(pressure.reshape(3, 3))
            if len(pressure) == 3:
                pressure = np.append(pressure, 3 * [None])
        self.input.minimizer = algorithm
        self.input.ionic_steps = max_iter
        self.input.pressure = pressure
        self.input.volume_only = volume_only
        self.input.ionic_force_tolerance = ionic_force_tolerance
        self.input.ionic_energy_tolerance = ionic_energy_tolerance
        self.input.pressure_tolerance = pressure_tolerance


class Input(DataContainer):
    """
    Args:
        minimizer (str): minimizer to use (currently only 'CG' and 'BFGS' run
            reliably)
        ionic_steps (int): max number of steps
        ionic_force_tolerance (float): maximum force tolerance
    """

    def __init__(self, input_file_name=None, table_name="input"):
        self.minimizer = "CG"
        self.ionic_steps = 100
        self.ionic_force_tolerance = 1.0e-2
        self.pressure = None
        self.volume_only = False
        self.ionic_energy_tolerance = 0
        self.pressure_tolerance = 1.0e-3


class ScipyMinimizerOutput(GenericInteractiveOutput):
    def __init__(self, job):
        super(ScipyMinimizerOutput, self).__init__(job=job)
        self._result = None

    def to_hdf(self, hdf, group_name="output"):
        if self._result is None:
            return
        with hdf.open(group_name) as hdf_output:
            hdf_output["convergence"] = self._result["success"]
            if "hess_inv" in self._result.keys():
                hdf_output["hessian"] = self._result["hess_inv"]



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import codecs
import pickle
from typing import Optional

import numpy as np
import posixpath
import scipy.constants
from phonopy import Phonopy
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.units import VaspToTHz
from phonopy.file_IO import write_FORCE_CONSTANTS

from pyiron_atomistics.atomistics.structure.atoms import Atoms
from pyiron_atomistics.atomistics.master.parallel import AtomisticParallelMaster
from pyiron_atomistics.atomistics.structure.phonopy import (
    publication as phonopy_publication,
)
from pyiron_base import state, JobGenerator, ImportAlarm, deprecate

__author__ = "Jan Janssen, Yury Lysogorskiy"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2017"


class Thermal:
    """
    Holds thermal properties that are the results of the phonopy calculation.

    Args:
        temperatures (ndarray): temperatures at which the other quantities are evaluated, units of Kelvin
        free_energies (ndarray): free energies in the quasi-harmonic approximation from phonon contributions, units of
                                 electron volts
        entropy (ndarray): vibrational entropy calculated from the above free energy
        cv (ndarray): heat capacity at constant volume, units of kJ/K/mol
    """

    def __init__(self, temperatures, free_energies, entropy, cv):

        self.temperatures = temperatures
        self.free_energies = free_energies
        self.entropy = entropy
        self.cv = cv


def phonopy_to_atoms(ph_atoms):
    """
    Convert Phonopy Atoms to ASE-like Atoms
    Args:
        ph_atoms: Phonopy Atoms object

    Returns: ASE-like Atoms object

    """
    return Atoms(
        symbols=list(ph_atoms.get_chemical_symbols()),
        positions=list(ph_atoms.get_positions()),
        cell=list(ph_atoms.get_cell()),
        pbc=True,
    )


def atoms_to_phonopy(atom):
    """
    Convert ASE-like Atoms to Phonopy Atoms
    Args:
        atom: ASE-like Atoms

    Returns:
        Phonopy Atoms

    """
    return PhonopyAtoms(
        symbols=list(atom.get_chemical_symbols()),
        scaled_positions=list(atom.get_scaled_positions()),
        cell=list(atom.get_cell()),
    )


class PhonopyJobGenerator(JobGenerator):
    @property
    def parameter_list(self):
        """

        Returns:
            (list)
        """
        supercells = self._master.phonopy.get_supercells_with_displacements()
        return [
            [
                "{}_{}".format(self._master.ref_job.job_name, ind),
                self._restore_magmoms(phonopy_to_atoms(sc)),
            ]
            for ind, sc in enumerate(supercells)
        ]

    def _restore_magmoms(self, structure):
        """
        Args:
            structure (pyiron.atomistics.structure.atoms): input structure

        Returns:
            structure (pyiron_atomistics.atomistics.structure.atoms): output structure with magnetic moments
        """
        if any(self._master.structure.get_initial_magnetic_moments() != None):
            magmoms = self._master.structure.get_initial_magnetic_moments()
            magmoms = np.tile(
                magmoms,
                np.prod(np.diagonal(self._master._phonopy_supercell_matrix())).astype(
                    int
                ),
            )
            structure.set_initial_magnetic_moments(magmoms)
        return structure

    @staticmethod
    def job_name(parameter):
        return parameter[0]

    def modify_job(self, job, parameter):
        job.structure = parameter[1]
        return job


class PhonopyJob(AtomisticParallelMaster):
    """
    Phonopy wrapper for the calculation of free energy in the framework of quasi harmonic approximation.

    Example:

    >>> from pyiron_atomistics import Project
    >>> pr = Project('my_project')
    >>> lmp = pr.create_job('Lammps', 'lammps')
    >>> lmp.structure = pr.create_structure('Fe', 'bcc', 2.832)
    >>> phono = lmp.create_job('PhonopyJob', 'phonopy')
    >>> phono.run()

    Get output via `get_thermal_properties()`.

    Note:

    - This class does not consider the thermal expansion. For this, use `QuasiHarmonicJob` (find more in its
        docstring)
    - Depending on the value given in `job.input['interaction_range']`, this class automatically changes the number of
        atoms. The input parameters of the reference job might have to be set appropriately (e.g. use `k_mesh_density`
        for DFT instead of setting k-points directly).
    - The structure used in the reference job should be a relaxed structure.
    - Theory behind it: https://en.wikipedia.org/wiki/Quasi-harmonic_approximation
    """

    def __init__(self, project, job_name):
        super(PhonopyJob, self).__init__(project, job_name)

        self.__version__ = "0.0.1"
        self.input["interaction_range"] = (10.0, "Minimal size of supercell, Ang")
        self.input["factor"] = (
            VaspToTHz,
            "Frequency unit conversion factor (default for VASP)",
        )
        self.input["displacement"] = (0.01, "atoms displacement, Ang")
        self.input["dos_mesh"] = (20, "mesh size for DOS calculation")
        self.input["primitive_matrix"] = None
        self.input["number_of_snapshots"] = (
            None,
            "int or None, optional. Number of snapshots of supercells with random displacements. Random displacements "
            "are generated displacing all atoms in random directions with a fixed displacement distance specified by "
            "'distance' parameter, i.e., all atoms in supercell are displaced with the same displacement distance in "
            "direct space. (Copied directly from phonopy docs. Requires the alm package to work.)",
        )

        self.phonopy = None
        self._job_generator = PhonopyJobGenerator(self)
        self._disable_phonopy_pickle = False
        state.publications.add(phonopy_publication())

    @property
    def phonopy_pickling_disabled(self):
        return self._disable_phonopy_pickle

    @phonopy_pickling_disabled.setter
    def phonopy_pickling_disabled(self, disable):
        self._disable_phonopy_pickle = disable

    @property
    def _phonopy_unit_cell(self):
        if self.structure is not None:
            return atoms_to_phonopy(self.structure)
        else:
            return None

    def _enable_phonopy(self):
        if self.phonopy is None:
            if self.structure is not None:
                self.phonopy = Phonopy(
                    unitcell=self._phonopy_unit_cell,
                    supercell_matrix=self._phonopy_supercell_matrix(),
                    primitive_matrix=self.input["primitive_matrix"],
                    factor=self.input["factor"],
                )
                self.phonopy.generate_displacements(
                    distance=self.input["displacement"],
                    number_of_snapshots=self.input["number_of_snapshots"],
                )
                self.to_hdf()
            else:
                raise ValueError("No reference job/ No reference structure found.")

    def list_structures(self):
        if self.structure is not None:
            self._enable_phonopy()
            return [struct for _, struct in self._job_generator.parameter_list]
        else:
            return []

    def _phonopy_supercell_matrix(self):
        if self.structure is not None:
            supercell_range = np.ceil(
                self.input["interaction_range"]
                / np.array(
                    [np.linalg.norm(vec) for vec in self._phonopy_unit_cell.get_cell()]
                )
            )
            return np.eye(3) * supercell_range
        else:
            return np.eye(3)

    def run_static(self):
        # Initialise the phonopy object before starting the first calculation.
        self._enable_phonopy()
        super(PhonopyJob, self).run_static()

    def run_if_interactive(self):
        self._enable_phonopy()
        super(PhonopyJob, self).run_if_interactive()

    def to_hdf(self, hdf=None, group_name=None):
        """
        Store the PhonopyJob in an HDF5 file

        Args:
            hdf (ProjectHDFio): HDF5 group object - optional
            group_name (str): HDF5 subgroup name - optional
        """
        super(PhonopyJob, self).to_hdf(hdf=hdf, group_name=group_name)
        if self.phonopy is not None and not self._disable_phonopy_pickle:
            with self.project_hdf5.open("output") as hdf5_output:
                hdf5_output["phonopy_pickeled"] = codecs.encode(
                    pickle.dumps(self.phonopy), "base64"
                ).decode()

    def from_hdf(self, hdf=None, group_name=None):
        """
        Restore the PhonopyJob from an HDF5 file

        Args:
            hdf (ProjectHDFio): HDF5 group object - optional
            group_name (str): HDF5 subgroup name - optional
        """
        super(PhonopyJob, self).from_hdf(hdf=hdf, group_name=group_name)
        with self.project_hdf5.open("output") as hdf5_output:
            if "phonopy_pickeled" in hdf5_output.list_nodes():
                self.phonopy = pickle.loads(
                    codecs.decode(hdf5_output["phonopy_pickeled"].encode(), "base64")
                )
                if "dos_total" in hdf5_output.list_nodes():
                    self._dos_total = hdf5_output["dos_total"]
                if "dos_energies" in hdf5_output.list_nodes():
                    self._dos_energies = hdf5_output["dos_energies"]

    def collect_output(self):
        """

        Returns:

        """
        if self.ref_job.server.run_mode.interactive:
            forces_lst = self.project_hdf5.inspect(self.child_ids[0])[
                "output/generic/forces"
            ]
        else:
            pr_job = self.project_hdf5.project.open(self.job_name + "_hdf5")
            forces_lst = [
                pr_job.inspect(job_name)["output/generic/forces"][-1]
                for job_name in self._get_jobs_sorted()
            ]
        self.phonopy.set_forces(forces_lst)
        self.phonopy.produce_force_constants(
            fc_calculator=None if self.input["number_of_snapshots"] is None else "alm"
        )
        self.phonopy.run_mesh(mesh=[self.input["dos_mesh"]] * 3)
        mesh_dict = self.phonopy.get_mesh_dict()
        self.phonopy.run_total_dos()
        dos_dict = self.phonopy.get_total_dos_dict()

        self.to_hdf()

        with self.project_hdf5.open("output") as hdf5_out:
            hdf5_out["dos_total"] = dos_dict["total_dos"]
            hdf5_out["dos_energies"] = dos_dict["frequency_points"]
            hdf5_out["qpoints"] = mesh_dict["qpoints"]
            hdf5_out["supercell_matrix"] = self._phonopy_supercell_matrix()
            hdf5_out["displacement_dataset"] = self.phonopy.get_displacement_dataset()
            hdf5_out[
                "dynamical_matrix"
            ] = self.phonopy.dynamical_matrix.get_dynamical_matrix()
            hdf5_out["force_constants"] = self.phonopy.force_constants

    def write_phonopy_force_constants(self, file_name="FORCE_CONSTANTS", cwd=None):
        """

        Args:
            file_name:
            cwd:

        Returns:

        """
        if cwd is not None:
            file_name = posixpath.join(cwd, file_name)
        write_FORCE_CONSTANTS(
            force_constants=self.phonopy.force_constants, filename=file_name
        )

    def get_hesse_matrix(self):
        """

        Returns:

        """
        unit_conversion = (
            scipy.constants.physical_constants["Hartree energy in eV"][0]
            / scipy.constants.physical_constants["Bohr radius"][0] ** 2
            * scipy.constants.angstrom**2
        )
        force_shape = np.shape(self.phonopy.force_constants)
        force_reshape = force_shape[0] * force_shape[2]
        return (
            np.transpose(self.phonopy.force_constants, (0, 2, 1, 3)).reshape(
                (force_reshape, force_reshape)
            )
            / unit_conversion
        )

    def get_thermal_properties(self, t_min=1, t_max=1500, t_step=50, temperatures=None):
        """
        Returns thermal properties at constant volume in the given temperature range.  Can only be called after job
        successfully ran.

        Args:
            t_min (float): minimum sample temperature
            t_max (float): maximum sample temperature
            t_step (int):  tempeature sample interval
            temperatures (array_like, float):  custom array of temperature samples, if given t_min, t_max, t_step are
                                               ignored.

        Returns:
            :class:`Thermal`: thermal properties as returned by Phonopy
        """
        self.phonopy.run_thermal_properties(
            t_step=t_step, t_max=t_max, t_min=t_min, temperatures=temperatures
        )
        tp_dict = self.phonopy.get_thermal_properties_dict()
        kJ_mol_to_eV = 1000 / scipy.constants.Avogadro / scipy.constants.electron_volt
        return Thermal(
            tp_dict["temperatures"],
            tp_dict["free_energy"] * kJ_mol_to_eV,
            tp_dict["entropy"],
            tp_dict["heat_capacity"],
        )

    @property
    def dos_total(self):
        """
        ndarray of float: Value of the DOS.
        """
        return self["output/dos_total"]

    @property
    def dos_energies(self):
        """
        ndarray of float: Energies at which the DOS is sampled.
        """
        return self["output/dos_energies"]

    @property
    def dynamical_matrix(self):
        """

        Returns:

        """
        return np.real_if_close(
            self.phonopy.get_dynamical_matrix().get_dynamical_matrix()
        )

    def dynamical_matrix_at_q(self, q):
        """

        Args:
            q:

        Returns:

        """
        return np.real_if_close(self.phonopy.get_dynamical_matrix_at_q(q))

    @deprecate(ax="Use axis instead")
    def plot_dos(self, ax=None, *args, axis=None, **kwargs):
        """
        Plot the DOS.

        If "label" is present in `kwargs` a legend is added to the plot automatically.

        Args:
            axis (optional): matplotlib axis to use, if None create a new one
            ax: deprecated alias for axis
            *args: passed to `axis.plot`
            **kwargs: passed to `axis.plot`

        Returns:
            matplotlib.axes._subplots.AxesSubplot: axis with the plot
        """
        try:
            import pylab as plt
        except ImportError:
            import matplotlib.pyplot as plt
        if ax is not None and axis is None:
            axis = ax
        if axis is None:
            _, axis = plt.subplots(1, 1)
        axis.plot(
            self["output/dos_energies"], self["output/dos_total"], *args, **kwargs
        )
        axis.set_xlabel("Frequency [THz]")
        axis.set_ylabel("DOS")
        axis.set_title("Phonon DOS vs Energy")
        if "label" in kwargs:
            axis.legend()
        return ax

    def get_band_structure(
        self, npoints=101, with_eigenvectors=False, with_group_velocities=False
    ):
        """
        Calculate band structure with automatic path through reciprocal space.

        Can only be called after job is finished.

        Args:
            npoints (int, optional):  Number of sample points between high symmetry points.
            with_eigenvectors (boolean, optional):  Calculate eigenvectors, too
            with_group_velocities (boolean, optional):  Calculate group velocities, too

        Returns:
            :class:`dict` of the results from phonopy under the following keys
                - 'qpoints':  list of (npoints, 3), samples paths in reciprocal space
                - 'distances':  list of (npoints,), distance along the paths in reciprocal space
                - 'frequencies':  list of (npoints, band), phonon frequencies
                - 'eigenvectors':  list of (npoints, band, band//3, 3), phonon eigenvectors
                - 'group_velocities': list of (npoints, band), group velocities
            where band is the number of bands (number of atoms * 3).  Each entry is a list of arrays, and each array
            corresponds to one path between two high-symmetry points automatically picked by Phonopy and may be of
            different length than other paths.  As compared to the phonopy output this method also reshapes the
            eigenvectors so that they directly have the same shape as the underlying structure.

        Raises:
            :exception:`ValueError`: method is called on a job that is not finished or aborted
        """
        if not self.status.finished:
            raise ValueError(
                "Job must be successfully run, before calling this method."
            )

        self.phonopy.auto_band_structure(
            npoints,
            with_eigenvectors=with_eigenvectors,
            with_group_velocities=with_group_velocities,
        )
        results = self.phonopy.get_band_structure_dict()
        if results["eigenvectors"] is not None:
            # see https://phonopy.github.io/phonopy/phonopy-module.html#eigenvectors for the way phonopy stores the
            # eigenvectors
            results["eigenvectors"] = [
                e.transpose(0, 2, 1).reshape(*e.shape[:2], -1, 3)
                for e in results["eigenvectors"]
            ]
        return results

    def plot_band_structure(
        self, axis=None, *args, label: Optional[str] = None, **kwargs
    ):
        """
        Plot bandstructure calculated with :method:`.get_bandstructure`.

        If :method:`.get_bandstructure` hasn't been called before, it is automatically called with the default arguments.

        If `label` is passed a legend is added automatically.

        Args:
            axis (matplotlib axis, optional): plot to this axis, if not given a new one is created.
            *args: passed through to matplotlib.pyplot.plot when plotting the dispersion
            label (str, optional): label for dispersion line
            **kwargs: passed through to matplotlib.pyplot.plot when plotting the dispersion

        Returns:
            matplib axis: the axis the figure has been drawn to, if axis is given the same object is returned
        """
        # label is it's own argument because if you try to pass it via **kwargs every line would get the label giving a
        # messy legend
        try:
            import pylab as plt
        except ImportError:
            import matplotlib.pyplot as plt

        if axis is None:
            _, axis = plt.subplots(1, 1)

        try:
            results = self.phonopy.get_band_structure_dict()
        except RuntimeError:
            results = self.get_band_structure()

        distances = results["distances"]
        frequencies = results["frequencies"]

        # HACK: strictly speaking this breaks phonopy API and could bite us
        path_connections = self.phonopy._band_structure.path_connections
        labels = self.phonopy._band_structure.labels

        if "color" not in kwargs:
            kwargs["color"] = "black"

        offset = 0
        tick_positions = [distances[0][0]]
        for di, fi, ci in zip(distances, frequencies, path_connections):
            axis.axvline(
                tick_positions[-1], color="black", linestyle="dotted", alpha=0.5
            )
            line, *_ = axis.plot(offset + di, fi, *args, **kwargs)
            tick_positions.append(di[-1] + offset)
            if not ci:
                offset += 0.05
                axis.axvline(
                    tick_positions[-1], color="black", linestyle="dotted", alpha=0.5
                )
                tick_positions.append(di[-1] + offset)
        if label is not None:
            line.set_label(label)
            axis.legend()
        axis.set_xticks(tick_positions[:-1])
        axis.set_xticklabels(labels)
        axis.set_xlabel("Bandpath")
        axis.set_ylabel("Frequency [THz]")
        axis.set_title("Bandstructure")
        return axis

    def validate_ready_to_run(self):
        if self.ref_job._generic_input["calc_mode"] != "static":
            raise ValueError(
                "Phonopy reference jobs should be static calculations, but got {}".format(
                    self.ref_job._generic_input["calc_mode"]
                )
            )
        if self.input["number_of_snapshots"] is not None:
            with ImportAlarm(
                "Phonopy with random snapshot displacement relies on the alm module but this unavailable. Please "
                "ensure your python environment contains alm, e.g. by running `conda install -c conda-forge alm`."
                "Note: at time of writing alm is only available for Linux and OSX systems."
            ) as import_alarm:
                import alm
            import_alarm.warn_if_failed()
        super().validate_ready_to_run()



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from pyiron_atomistics.lammps.interactive import LammpsInteractive

__author__ = "Joerg Neugebauer, Sudarsan Surendralal, Jan Janssen"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "- Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Sudarsan Surendralal"
__email__ = "surendralal@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"


class Lammps(LammpsInteractive):
    """
    Class to setup and run and analyze LAMMPS simulations.

    Example:

    >>> job = pr.create.job.Lammps(job_name='lmp_example')
    >>> job.structure = pr.create.structure.bulk('Fe', cubic=True)
    >>> job.run()

    How to set potential: Look up potentials via `job.view_potentials()` (detailed data frame)
    or via `job.list_potentials()` (potential names). Assign the potential e.g. via:

    >>> job.potential = job.list_potentials()[0]

    Lammps has 3 modes: `static`, `md` and `minimize`. Set a mode e.g. via:

    >>> job.calc_minimize()

    Args:
        project (pyiron_atomistics.project.Project instance):  Specifies the project path among other attributes
        job_name (str): Name of the job

    Attributes:
        input (lammps.Input instance): Instance which handles the input
    """

    def __init__(self, project, job_name):
        super(Lammps, self).__init__(project, job_name)

        self._executable_activate(enforce=True)
